#https://midgard.thorchain.info/v2/actions?offset=0&limit=10&address=thor1khjg7et54q4zzef62ud0z8ps0ypnf69g5hfxjc

#https://thornode.thorchain.info/bank/balances/thor1khjg7et54q4zzef62ud0z8ps0ypnf69g5hfxjc

#https://midgard.thorchain.info/v2/member/thor1khjg7et54q4zzef62ud0z8ps0ypnf69g5hfxjc --> all pools
#https://viewblock.io/thorchain/address/thor12dpps82a5a2zmzdvq74rjunfjh2szykgcr9anw?page=1

from datetime import datetime
import pytz as timezone
from decimal import Decimal
from copy import deepcopy

from pyCryptoTransactions.Transaction import Position, Transaction,TransactionList, Fee
from pyCryptoTransactions.Importer import Importer
import time
import datetime
import requests
from requests.exceptions import ConnectionError, Timeout, TooManyRedirects
import re, json

class ThorchainViewBlockImporter(Importer):
    def __init__(self, address):
        super().__init__()
        self.address = address
        self._apiAdress = 'https://viewblock.io/thorchain/address/' + self.address
        #self._initRequest()

        self.rawTxs = None

        self.denominator = Decimal(int(1E8))
        self.chain = "thorchain"
        self.unit = "RUNE"
    
    def _request(self, path, apiAdress=None, **params):
        if apiAdress == None:
                apiAdress = self._apiAdress
        try:
            # NOTE: Session doesn't work here, some parts are missing
            response = requests.get(apiAdress + path)
            #response = self.session.get(apiAdress + path, params=params)
            if response.status_code != 200:
                raise("Error in the request")
            text = response.text
            r = re.search('window\.\_\_INITIAL\_STATE\_\_ =(.+)', text)
            d = json.loads(r.group(1))
            return d["main"]["address"]["thorchain"]["map"][self.address]
        except (ConnectionError, Timeout, TooManyRedirects) as e:
            print(e)
            raise("Connection Error")
    
    def _getRawTransactions(self, page=0, **kwargs):
        values = {}
        values['page'] = page
        for k in kwargs:
            if kwargs[k] is not None:
                values[k] = kwargs[k]
        return self._request('/', **values)

    def _getAllRawTransactions(self):
        return NotImplemented

    def getTransactions(self, time=None) -> TransactionList:
        txsRaw = self._getRawTransactions() #todo:replace with getAllRawTransactions
        for tx in txsRaw["txs"]["docs"]:
            time = datetime.datetime.fromtimestamp(int(tx["timestamp"])/1E3, tz=timezone.utc)
            txHash = tx["hash"]
            fee = Fee( Decimal(tx["fee"])/self.denominator, self.unit)
            memo = '_'.join(tx["extra"]["thorMemos"])
            category = '_'.join(tx["extra"]["thorLabels"])
            #status = tx["status"] == "success"
            t = Transaction(dateTime=time, fee=fee, txHash=txHash, category=category)
            t.memo = memo

            outgoing = False
            ingoing = False
            for transfer in tx["transfers"]:
                amount = Decimal(int(0))
                for fromTransfer in transfer["from"]:
                    if fromTransfer["address"] == self.address:
                        outgoing = True
                        t.fromAddress = self.address
                    t.fromAddress = fromTransfer["address"] if t.fromAddress is not None or t.fromAddress != "" else t.fromAddress
                        # amount seems to be always in the "to" transfer --> TODO: Check if this is alway true!
                for toTransfer in transfer["to"]:
                    if toTransfer["address"] == self.address:
                        ingoing = True
                        t.toAddress = self.address
                    t.toAddress = fromTransfer["address"] if t.toAddress is not None or t.toAddress != "" else t.toAddress
                    if amount != 0:
                        raise ValueError("Amount was already set once! Check implementation and API (changes)")
                    amount += Decimal(toTransfer["value"])/self.denominator
                
            # Thorchain is just RUNE for now, so there can't be in and out rune to our address at the same time
            if outgoing and ingoing:
                raise Exception("Both in and out going transaction detected! Check implementation and API (changes)!")
            
            if outgoing:
                t.posOut = Position(amount, "RUNE")
            elif ingoing:
                t.posIn = Position(amount, "RUNE")
                t.fee.amount = 0
            else:
                print("Neither in nor out transaction detected!")
            
            #if tx["extra"]["thorOps"]["hasOps"]:
            #    if ingoing:
            #        t.blockHeight = tx["extra"]["thorOps"]["outs"][0]["blockHeight"] #thorchain outs are our ins #TODO: can there be more than 1?
            #    elif outgoing:
            #        t.blockHeight = tx["extra"]["thorOps"]["ins"][0]["blockHeight"]
                
            self.txList.append(t)
        return self.txList



from typing import Optional
from pandas.core.frame import DataFrame
from pyCryptoTransactions.Transaction import TransactionList
from pyCryptoTransactions.exporters.default import ExporterDefault
import datetime

class AccointingExporter(ExporterDefault):

    def __init__(self, txData: TransactionList) -> None:
        super().__init__(txData)
        self.parseTxs()

    def parseTxs(self):
        data = []
        for tx in self.txData:
            feeCost = False
            if tx.posIn.amount != 0 and tx.posOut.amount != 0:
                tType = "order"
            elif tx.posIn.amount != 0 and tx.posOut.amount == 0:
                tType = "deposit"
            elif tx.posIn.amount == 0 and tx.posOut.amount != 0:
                tType = "withdraw"
            elif tx.posIn.amount == 0 and tx.posOut.amount == 0 and tx.fee.amount != 0:
                feeCost = True
                tType = "withdraw"
            else:
                print("Accointing Export: Unknown type: Tx {}".format(tx.txHash))
                print(tx)
                continue
            
            if tx.category.lower() == "mining":
                classification = "mined"
            elif tx.category.lower() == "income":
                classification = "income"
            elif tx.category.lower() == "airdrop":
                classification = "airdrop" 
            elif tx.category.lower() == "lending income":
                classification = "lending_income"
            elif tx.category.lower() == "lending":
                classification = "lending"
            elif tx.category.lower() == "staking income" or tx.category.lower() == "staking":
                classification = "staked"
            elif tx.category.lower() == "bounty":
                classification = "bounty"
            else:
                classification = ""

            entry = {"transactionType":tType,
                    "date":datetime.datetime.strftime(tx.datetime, "%m/%d/%Y %H:%M:%S"),
                    "inBuyAmount": tx.posIn.amount if tx.posIn.amount != 0 else "",
                    "inBuyAsset": tx.posIn.currency if tx.posIn.amount != 0 else "",

                    "outSellAmount": tx.fee.amount if feeCost else (tx.posOut.amount if tx.posOut.amount != 0 else ""),
                    "outSellAsset": tx.fee.currency if feeCost else (tx.posOut.currency if tx.posOut.amount != 0 else ""),
                    "feeAmount (optional)": tx.fee.amount if tx.fee.amount >0 and not feeCost else "",
                    "feeAsset (optional)": tx.fee.currency if tx.fee.amount > 0 and not feeCost else "",

                    "classification (optional)": "fee" if feeCost else classification, #no support for now
                    "operationId (optional)": tx.txHash}
            data.append(entry)
        self.df = DataFrame(data)
    




#transactionType	date	inBuyAmount	inBuyAsset	outSellAmount	outSellAsset	feeAmount (optional)	feeAsset (optional)	classification (optional)	operationId (optional)

#intags:
#add funds	airdrop	bounty	
#gambling_income
#gift_received
#hard_fork	ignored	 
#income	lending_income	
#liquidity_pool
#margin_gain	master_node	mined	staked

#outTags:
#remove funds	payment	fee	gambling_used	gift_sent	ignored	interest_paid	 
#lending	lost	margin_fee	margin_loss	payment

#MM/DD/YYYY HH:mm:SS


from abc import abstractclassmethod, abstractmethod
from requests.exceptions import ConnectionError, Timeout, TooManyRedirects
from requests import Request, Session
import json
import time

import abc
from pyCryptoTransactions.Transaction import TransactionList,Transaction
#import Transaction, TransactionList

class Importer(metaclass=abc.ABCMeta):

    Network = None

    def __init__(self):
        self.txList = TransactionList()
        self._apiAdress = ""

    #@property
    #def BLOCK_EXPLORER_URL(self):
    #    return self._blockExplorerUrl

    @abc.abstractmethod
    def getTransactions(self, time=None) -> TransactionList:
        pass

    ########## HELPER FUNCTION ######################

    def _returnValueFromPath(self, dictElement, path):
        mydata = dictElement
        for i in path:
            mydata = mydata[i]
        return mydata
    
    def _initRequest(self):
        headers = {'Accepts': 'application/json'}
        self.session = Session()
        self.session.headers.update(headers)
    
    def _request(self, path, apiAdress = None, **params):
        if apiAdress == None:
            apiAdress = self._apiAdress
        try:
            #print(params)
            response = self.session.get(apiAdress + path, params=params)
            #print(response.url)
            if response.status_code != 200:
                raise("Error in the request")

            data = json.loads(response.text)
            if "ok" in data:
                if not(data["ok"]):
                    print(data)
                    raise("Error in the Request")
            if "code" in data:
                if data["code"] != 0:
                    print(data)
                    raise("Error in the request")
            if data: #data['ok'] and data["code"]==0:
                time.sleep(1)
                return data
            else:
                print("No data in request") #raise("Data failure")
        except (ConnectionError, Timeout, TooManyRedirects) as e:
            print(e)
            raise("Connection Error")


from numpy import True_, inf
from pandas.core.frame import DataFrame
from pyCryptoTransactions.Importer import Importer
from pyCryptoTransactions.Transaction import Position, Transaction, TransactionList, Fee
from decimal import Decimal
import pandas as pd
import pytz
import datetime
import re
from copy import deepcopy

local = pytz.timezone("Europe/Berlin")
localUTC = pytz.utc

class BinanceManual(Importer):

    def __init__(self):
        super().__init__()
        self._apiAdress = "https://api.binance.com/api/v3/exchangeInfo"
        #https://api.binance.com/api/v3/exchangeInfo?symbol=BNBBTC
        self.exchangeInfo = None
        self._initRequest()
        self._getExchangeInfo()
    
    def _getExchangeInfo(self):
        data= self._request('')
        self.exchangeInfo = data

    def parseBuyHistory(self, excelFile):
        pass

    def _loadExcelFile(self, excelFile):
        df = pd.read_excel(excelFile)
        headers = list(df)
        return df, headers

    def parseDepositCryptoHistory(self, excelFile) -> TransactionList:
        return self.__parseDepsitWithdrawalCryptoHistory(excelFile, deposit=True)
    
    def parseWithdrawalCryptoHistory(self, excelFile) -> TransactionList:
        return self.__parseDepsitWithdrawalCryptoHistory(excelFile, deposit=False)

    def __parseDepsitWithdrawalCryptoHistory(self, excelFile, deposit:bool=True) -> TransactionList:
        df, headers = self._loadExcelFile(excelFile)
        tmpList = TransactionList()

        for index, row in df.iterrows():
            if not(self._checkStatus(row["Status"])):
                continue
            time = self._getDatetime(row[headers[0]], headers[0])
            if deposit:
                t = Transaction(dateTime=time, 
                                posIn=Position( Decimal(str(row["Amount"])), row["Coin"] ),
                                txHash=row["TXID"])
                t.fromAddress = row["Address"]
            else:
                t = Transaction(dateTime=time, 
                                posOut=Position( Decimal(str(row["Amount"])), row["Coin"] ),
                                fee = Fee( Decimal(str(row["TransactionFee"])), row["Coin"]),
                                txHash=row["TXID"])
                t.toAddress = row["Address"]
            tmpList.append(t)

        self.txList += tmpList
        return tmpList
    
    def parseDepositFiatHistory(self, excelFile) -> TransactionList:
        df, headers = self._loadExcelFile(excelFile)
        tmpList = TransactionList()

        for index, row in df.iterrows():
            if not(self._checkStatus(row["Status"])):
                continue
            time = self._getDatetime(row[headers[0]], headers[0])
            t = Transaction(dateTime=time, 
                            posIn=Position( Decimal(str(row["Indicated Amount"])), row["Coin"] ),
                            fee=Fee( Decimal(str(row["Fee"])), row["Coin"] ))
            t.orderId = row["Order ID"]
            t.note = "Deposit by " + row["Payment Method"]
            tmpList.append(t)

        self.txList += tmpList
        return tmpList

    def parseBuyCryptoHistory(self, excelFile) -> TransactionList:
        df, headers = self._loadExcelFile(excelFile)
        tmpList = TransactionList()

        for index, row in df.iterrows():
            if not(self._checkStatus(row["Status"])):
                continue
            time = self._getDatetime(row[headers[0]], headers[0])

            amountOut, currencyOut = self._getAmountCurrency(str(row["Amount"]))
            amountIn, currencyIn = self._getAmountCurrency(str(row["Final Amount"]))
            fee, feeCurrency = self._getAmountCurrency(str(row["Fees"]))

            if row["Method"] != "Cash Balance":
                # Make deposit
                tDeposit = Transaction(dateTime=time, posIn=Position( amountOut, currencyOut) )
                tDeposit.note = "Deposit for Crypto Buy with " + row["Method"]
                tmpList.append(tDeposit)

            t = Transaction(dateTime=time, 
                            posIn=Position( amountIn, currencyIn),
                            posOut=Position( amountOut - fee if currencyOut==feeCurrency else amountOut, currencyOut),
                            fee=Fee( fee, feeCurrency))
            t.orderId = row["Transaction ID"]
            tmpList.append(t)

        self.txList += tmpList
        return tmpList
    
    def parseTradeHistory(self, excelFile) -> TransactionList:
        df, headers = self._loadExcelFile(excelFile)
        tmpList = TransactionList()
        for index, row in df.iterrows():
            t = Transaction(dateTime=self._getDatetime(row[headers[0]], headers[0]))
            market = row["Market"]
            baseAsset, quoteAsset = self._getAssetsFromMarket(market)
            amount = Decimal( str(row["Amount"]) )
            total = Decimal( str(row["Total"]) )
            if row["Type"].lower() == "buy":
                posIn = Position( amount, baseAsset )
                posOut = Position( total, quoteAsset )
            elif row["Type"].lower() == "sell":
                posOut = Position( amount, baseAsset )
                posIn = Position( total, quoteAsset )
            else:
                continue
            t.posIn = posIn
            t.posOut = posOut
            t.fee = Fee( Decimal(str(row["Fee"])), row["Fee Coin"] )
            tmpList.append(t)

        self.txList += tmpList
        return tmpList

        #tpye = buy --> posIn
        #market = LUNAEUR ---> try to split ...  --> LUNA EUR 
        # BUY: posIn = "Amount" LUNA, posOut = "Total" EUR
        # SELL: posOut = "Amount" LUNA, posIn = "TOTAL" EUR 
        # +Fee
    
    def _getAssetsFromMarket(self, market:str):
        if self.exchangeInfo is None:
            raise BaseException("Exchange Info to match markets is not available")
        
        if not "symbols" in self.exchangeInfo:
            raise BaseException("No symbols found in exchange info")
        
        for info in self.exchangeInfo["symbols"]:
            if info["symbol"] == market:
                baseAsset = info["baseAsset"]
                quoteAsset = info["quoteAsset"]
                return baseAsset,quoteAsset
        return None

    
    def parseStatements(self, csvFile) -> TransactionList:
        df = self._loadStatementsCsv(csvFile)
        self.txList += self._parseStatementBNBVault(df)
        self.txList += self._parseStatementLaunchpoolInterest(df)
        self.txList += self._parseStatementSavingsInterest(df)
        self.txList += self._parseStatementPosSavingsInterest(df)
        self.txList += self._parseStatementsCommision(df)
        self.txList += self._parseStatementLiquiditySwapRewards(df)
        self.txList += self._parseStatementLiquidtiyAdd(df)
        self.txList += self._parseStatementDistribution(df)
        self.txList += self._parseStatementETH2StakingRewards(df)
        self.txList += self._parseStatementETH2Staking(df)
        self.txList += self._parseStatementLaunchpadTokenDistribution(df)
        self.txList += self._parseStatementLaunchpadSubscribe(df)
        

    def _parseStatementTrades(self, csvFile):
        n = pd.read_csv(csvFile,sep=";")
        tmp = n.copy()
        tmp.UTC_Time = tmp.UTC_Time.dt.round("10s")
        #print(tmp)
        uniqueDates = tmp.UTC_Time.unique()
        for date in uniqueDates:
            #print(date)
            self.__parsteTradesFromStatements(tmp.loc[tmp['UTC_Time'] == date])

    def __parsteTradesFromStatements(self, dataFrame):
        feeCoins =  dataFrame.loc[dataFrame['Operation'] == 'Fee', 'Coin'].unique()
        buyCoins = dataFrame.loc[dataFrame['Operation'] == 'Buy', 'Coin'].unique()
        sellCoins = dataFrame.loc[dataFrame['Operation'] == 'Sell', 'Coin'].unique()
        transactionRelatedCoins = dataFrame.loc[dataFrame['Operation'] == 'Transaction Related', 'Coin'].unique()
        
        # Calculate Fee
        if len(feeCoins)>1:
            print("ERROR: MORE THAN 1 fee Units")
            print(dataFrame.UTC_Time.unique())
        feeCoin = feeCoins[0] if feeCoins.size>0 else None
        totalFee = dataFrame.loc[dataFrame['Operation'] == 'Fee', 'Change'].sum()
        
        # Get all Buy Coins
        b = {}
        for coin in buyCoins:
            b[coin] = dataFrame.loc[(dataFrame['Operation'] == 'Buy') & (dataFrame['Coin'] == coin), 'Change'].sum()
            
        # Get all sell coins
        s = {}
        for coin in sellCoins:
            s[coin] = dataFrame.loc[(dataFrame['Operation'] == 'Sell') & (dataFrame['Coin'] == coin), 'Change'].sum()
        
        # Get all transaction related coins
        t = {}
        for coin in transactionRelatedCoins:
            t[coin] = dataFrame.loc[(dataFrame['Operation'] == 'Transaction Related') & (dataFrame['Coin'] == coin), 'Change'].sum()
        
        # Create Trades
        if len(t) == 1 and len(t)==1 and t[coin]<0 and b[buyCoins[0]]>0 and len(s)==0:
            totalBuy = b[buyCoins[0]]
            print("Trade: {:.4f}{} for {:.4f}{} (fee {:.4f}{})".format(totalBuy, buyCoins[0], t[coin], coin, totalFee, feeCoin))
        elif len(t) == 2 and buyCoins.size==0:
            coin1 = transactionRelatedCoins[0]
            coin2 = transactionRelatedCoins[1]
            buyCoin = coin1 if t[coin1] > 0 else coin2
            sellCoin = coin1 if t[coin1] < 0 else coin2
            if buyCoin == sellCoin:
                print("ERROR: SAME COIN IN TRADE")
            totalBuy = t[buyCoin]
            totalSell = t[sellCoin]
            print("Trade: {:.4f}{} for {:.4f}{} (fee {:.4f}{})".format(totalBuy, buyCoin, totalSell, sellCoin, totalFee, feeCoin))
        elif len(t) == 0 and buyCoins.size==2:
            coin1 = buyCoins[0]
            coin2 = buyCoins[1]
            buyCoin = coin1 if b[coin1] > 0 else coin2
            sellCoin = coin1 if b[coin1] < 0 else coin2
            if buyCoin == sellCoin:
                print("ERROR: SAME COIN IN TRADE")
            totalBuy = b[buyCoin]
            totalSell = b[sellCoin]
            print("Trade: {:.4f}{} for {:.4f}{} (fee {:.4f}{})".format(totalBuy, buyCoin, totalSell, sellCoin, totalFee, feeCoin))
        elif len(s) == 2 and len(t) == 0 and len(b) == 0:
            coin1 = sellCoins[0]
            coin2 = sellCoins[1]
            buyCoin = coin1 if s[coin1] > 0 else coin2
            sellCoin = coin1 if s[coin1] < 0 else coin2
            if buyCoin == sellCoin:
                print("ERROR: SAME COIN IN TRADE")
            totalBuy = s[buyCoin]
            totalSell = s[sellCoin]
            print("Trade: {:.4f}{} for {:.4f}{} (fee {:.4f}{})".format(totalBuy, buyCoin, totalSell, sellCoin, totalFee, feeCoin))
        elif len(t) == 0 and buyCoins.size==0:
            foo = 1
        else:
            print("NOT RECOGNIZED")
            print(len(t))
            print(len(b))
            print(b[buyCoins[0]])
            print(dataFrame.UTC_Time.unique())
            
        # todo: 2 buys (- and +)
        # todo: 2 sells (- and +)

    def _parseStatementBNBSmallAssetConvert(self, csvFile):
        df = pd.read_csv(csvFile,sep=";")
        df = df.loc[df['Operation'] == 'Small Assets Exchange BNB']
        return NotImplemented
    
    def _parseStatementBNBVault(self, df:DataFrame) -> TransactionList:
        return self.__parseStatementInOrOut(df, "Super BNB Mining", "Mining")
    
    def _parseStatementSavingsInterest(self, df:DataFrame) -> TransactionList:
        return self.__parseStatementInOrOut(df, "Savings Interest", "Income")
    
    def _parseStatementPosSavingsInterest(self, df:DataFrame) -> TransactionList:
        return self.__parseStatementInOrOut(df, "POS Savings Interest", "Income")
    
    def _parseStatementsCommision(self, df:DataFrame) -> TransactionList:
        return self.__parseStatementInOrOut(df, "Commission History", "Income")
    
    def _parseStatementLaunchpoolInterest(self, df:DataFrame) -> TransactionList:
        return self.__parseStatementInOrOut(df, "Launchpool Interest", "Airdrop")
    
    def _parseStatementLiquiditySwapRewards(self, df:DataFrame) -> TransactionList:
        return self.__parseStatementInOrOut(df, "Liquid Swap rewards", "Lending Income")

    def _parseStatementDistribution(self, df:DataFrame) -> TransactionList:
        return self.__parseStatementInOrOut(df, "Distribution", "Income") # or bounty? #airdrop?

    def _parseStatementETH2StakingRewards(self, df:DataFrame) -> TransactionList:
        return self.__parseStatementInOrOut(df, "ETH 2.0 Staking Rewards", "Staking Income")

    def _parseStatementLiquidtiyAdd(self, df:DataFrame) -> TransactionList:
        return self.__parseStatementInOrOut(df, "Liquid Swap add", "Lending", inTx=False) 

    def _parseStatementLaunchpadTokenDistribution(self, df:DataFrame) -> TransactionList:
        return self.__parseStatementInOrOut(df, "Launchpad token distribution", "Bounty")
    
    def _parseStatementLaunchpadSubscribe(self, df:DataFrame) -> TransactionList:
        insAndOuts = self.__parseStatementInOrOut(df, "Launchpad subscribe", "")
        #return insAndOuts
        tmpList = TransactionList()
        for tx in insAndOuts:
            tx2 = deepcopy(tx)
            if tx2.posIn.amount < 0:
                tx2.posOut.amount = abs(tx2.posIn.amount)
                tx2.posOut.currency = tx2.posIn.currency
                tx2.posIn.amount = 0
                tx2.note = "Launchpad"
            tmpList.append(tx2)

                #todo: category
        return tmpList
    
    def _parseStatementETH2Staking(self, df:DataFrame) -> TransactionList:
        insAndOuts = self.__parseStatementInOrOut(df, "ETH 2.0 Staking", "")
        tmpList = TransactionList()
        for inTx in insAndOuts:
            for outTx in insAndOuts:
                if inTx.datetime == outTx.datetime and inTx.posIn.amount > 0 and outTx.posIn.amount < 0:
                    t = Transaction(dateTime=inTx.datetime, posIn=inTx.posIn, posOut=outTx.posOut)
                    t.note = "ETH 2.0 Staking"
                    insAndOuts.remove(inTx)
                    insAndOuts.remove(outTx)
                    tmpList.append(t)
        if insAndOuts:
            raise Exception("parseStatementETH2Staking: Not all ETH 2 staking transactions could be evaluated!")
        return tmpList

    
    def __parseStatementInOrOut(self, dataFrame:DataFrame, operationName:str, setCategory:str="", inTx:bool=True) -> TransactionList:
        """Simple in or out txs without fee"""
        df = dataFrame.copy()
        headers = list(df)
        v2 = False
        if headers[0] == "User_ID":
            v2 = True
        df = df.loc[df['Operation'] == operationName]
        txs = TransactionList()
        for idx, row in df.iterrows():
            time = self._getDatetime(row[headers[1] if v2 else headers[0]], headers[1] if v2 else headers[0])
            if inTx:
                t = Transaction(dateTime=time, posIn=Position( Decimal( str(row["Change"]) ), row["Coin"]))
            else:
                t = Transaction(dateTime=time, posOut=Position( abs(Decimal( str(row["Change"]) )), row["Coin"]))
            t.category = setCategory
            txs.append(t)
        
        return txs
    
    def _loadStatementsCsv(self, csvFile) -> DataFrame:
        df = pd.read_csv(csvFile,sep=";")
        # newer versions of the statements include the user id and use , as separator
        if len(df.columns) == 1:
            df = pd.read_csv(csvFile,sep=",")
            assert(len(df.columns) > 1)
        return df

    def _getDatetime(self, dateAsString:str, csvRowName:str) -> datetime.datetime:
        if not "date" in csvRowName.lower() and not "utc_time" in csvRowName.lower():
            raise BaseException("Non parseable header in Excel file!")
        
        # try to get UTC+1 information
        r = re.compile(r"Date\(UTC\+(\d)\)")
        m = r.match(csvRowName)
        if m:
            deltaHours = int(m.group(1))
        else:
            #print("not found")
            #print(csvRowName)
            deltaHours = 0
        
        d = datetime.datetime.fromisoformat(dateAsString)
        d = d - datetime.timedelta(hours=deltaHours)
        return d

    def _getAmountCurrency(self, amountStr:str):
        """Assuming 1234.56 COIN"""
        amount = Decimal(amountStr.split(" ")[0])
        currency = amountStr.split(" ")[1]
        return amount, currency
    
    def _checkStatus(self, status:str) -> bool:
        if status != "Successful" and status != "Completed":
            return False
        return True


    def getTransactions(self, time=None) -> TransactionList:
        self.txList.sort(key=lambda r: r.datetime)
        return self.txList


#POS savings purchase = Staking In               --------     wird vernachlässigt
#POS savings redemption = Staking Out            -------      dito
#POS savings interest = Staking Interest         

#Savings purchase                               -------- vernachlässigt
#Savings Principal redemption (LDBNB to BNB)    ######
#Savings Interest

#Commission History

#Super BNB Mining = Mining
#Launchpool Interest = Airdrop ?

#Liquid Swap add

#Small assets exchange BNB

#ETH 2.0 Staking
#Launchpad token distribution
#Liquid Swap rewards


import json
import urllib.request

class Currency(object):
    """
    This is a generic class for representing a currency,
    """

    @property
    def symbol(self):
        """
        The symbol name of the coin which is a short form of the coin name.
        Returns:
            The symbol name of the coin
        """
        return self.__symbol

    @property
    def name(self):
        """
        The name of the coin.
        Returns:
            The full name of the coin
        """
        return self.__name

    def __init__(self, name, symbol, **kwargs):
        super().__init__()

        self.__name = name
        self.__symbol = symbol

    def __str__(self):
        return self.__symbol


class CryptoCurrency(Currency):
    """
    This is a special type of currency. In order to initialize a CrpytoCurrency, use can use the `CryptoList` which will
    query coinmarketcap in order to retrieve a list of all available cryptocurrency coins/tokens.
    """

    def __init__(self, id, **kwargs):
        super().__init__(**kwargs)

        self.__id = id

    @property
    def id(self):
        """
        The id of the coin given by the coinmarketcap API.
        Returns:
            The id of the coin
        """
        return self.__id


class CryptoList(list):
    """
    This class is a list of all available crypto coins with each coin represented by its symbol name.
    On initialization, the list of crypto coins will be queried using the APIv2 of coinmarketcap.
    """

    _COINTMARKETCAP_QUERY_LISTING = 'https://api.coinmarketcap.com/v2/listings/'

    def __query_coinmarketcap(self):

        with urllib.request.urlopen(self._COINTMARKETCAP_QUERY_LISTING) as response:
            data = json.loads(response.read().decode())

            if data:
                return data['data']

    def __init__(self):

        super().__init__()

        self._coin_map = {}

        for entry in self.__query_coinmarketcap():
            c = CryptoCurrency(**entry)

            self.append(c)

            self._coin_map[c.symbol] = c

    def find_symbol(self, symbol):

        if symbol not in self._coin_map:

            if symbol == "IOTA":
                return self._coin_map["MIOTA"]

        else:
            return self._coin_map[symbol]


#!/usr/bin/env python
# https://en.bitcoin.it/wiki/Protocol_documentation#Addresses

import hashlib
import base58

# ECDSA bitcoin Public Key
pubkey = '0450863ad64a87ae8a2fe83c1af1a8403cb53f53e486d8511dad8a04887e5b23522cd470243453a299fa9e77237716103abc11a1df38855ed6f2ee187e9c582ba6'
#pubkey = 'xpub6C2zcaw8NCGbNwEo5KGPBYzNj6UpfspemmAfgsgxWTRvNPrXR6Sh9pvZvfNTjHaZ8mVPZjxMgi8RbCgecGSecU64agjh9uxQMXwPkxPf9bP'
# See 'compressed form' at https://en.bitcoin.it/wiki/Protocol_documentation#Signatures
compress_pubkey = False


def hash160(hex_str):
    sha = hashlib.sha256()
    rip = hashlib.new('ripemd160')
    sha.update(hex_str)
    rip.update( sha.digest() )
    print ( "key_hash = \t" + rip.hexdigest() )
    return rip.hexdigest()  # .hexdigest() is hex ASCII


if (compress_pubkey):
    if (ord(bytearray.fromhex(pubkey[-2:])) % 2 == 0):
        pubkey_compressed = '02'
    else:
        pubkey_compressed = '03'
    pubkey_compressed += pubkey[2:66]
    hex_str = bytearray.fromhex(pubkey_compressed)
else:
    hex_str = bytearray.fromhex(pubkey)

# Obtain key:

key_hash = '00' + hash160(hex_str)

# Obtain signature:

sha = hashlib.sha256()
sha.update( bytearray.fromhex(key_hash) )
checksum = sha.digest()
sha = hashlib.sha256()
sha.update(checksum)
checksum = sha.hexdigest()[0:8]

print ( "checksum = \t" + sha.hexdigest() )
print ( "key_hash + checksum = \t" + key_hash + ' ' + checksum )
print ( "bitcoin address = \t" + (base58.b58encode( bytes(bytearray.fromhex(key_hash + checksum)) )).decode('utf-8') )


from pycoin.symbols.btc import network as BTC
from pycoin.convention import satoshi_to_btc
import requests
import time
import blockcypher
from enum import Enum

from pyCryptoTransactions.networks.Network import Network
from pyCryptoTransactions.Transaction import Position, Transaction, TransactionList, Fee
from pyCryptoTransactions.Importer import Importer

class AddresType(Enum):
    LEGACY = 0
    SEGWIT = 1
    NATIVE_SEGWIT = 2
    

class BitcoinImport(Importer):
    Network = BTC

    def __init__(self, publicKey):
        '''
        The __init__ function is called when an instance of the class is created
        :param publicKey: The public key of the user
        '''
        super().__init__()
        self.publicKey = publicKey
        self._blockExplorerUrl = "http://blockchain.info"
    
    def _getSubkey(self, receiveAddr:bool=True, changeAddr:bool=False, number=0):
        '''
        It takes the public key and the desired subkey path as arguments and returns the subkey
        
        :param receiveAddr: True if you want to receive coins, False if you want to change coins, defaults
        to True
        :type receiveAddr: bool (optional)
        :param changeAddr: If True, return the change address for the account. If False, return the
        receiving address, defaults to False
        :type changeAddr: bool (optional)
        :param number: the index of the subkey to retrieve, defaults to 0 (optional)
        :return: The subkey object.
        '''
        if receiveAddr == changeAddr:
            raise ValueError("Either choose receiving address or change address, not both!")

        key = self.Network.parse(self.publicKey) #direct version (xpub=legacy, ypub=segwit, zpub=native segwit)
        subkey = key.subkey_for_path("{0}/{1}".format("0" if receiveAddr else "1", str(number)))
        return subkey

    def generateAddresses(self, func, receiveAddr:bool=True, changeAddr:bool=False, numAddresses=1000):
        '''
        Generate a list of addresses
        
        :param func: the function that generates the address
        :param receiveAddr: if True, generate a receive address. If False, then generate a change
        address, defaults to True
        :type receiveAddr: bool (optional)
        :param changeAddr: if True, generate a change address, defaults to False
        :type changeAddr: bool (optional)
        :param numAddresses: the number of addresses to generate, defaults to 1000 (optional)
        :return: A generator object.
        '''
        # returns a generator object
        counter = 0
        while True:
            if (counter > numAddresses): 
                return
            yield func(receiveAddr, changeAddr, counter)
            counter += 1

    def generateAddress(self, receiveAddr:bool=True, changeAddr:bool=False, number=0):
        '''
        If you want to generate an address, call this function
        
        :param receiveAddr: If True, the address will be marked as a receiving address. If False, it will be
        marked as a change address, defaults to True
        :type receiveAddr: bool (optional)
        :param changeAddr: If True, the address will be for change (rather than receive), defaults to False
        :type changeAddr: bool (optional)
        :param number: The index of the subkey to generate, defaults to 0 (optional)
        :return: The address of the subkey.
        '''
        subkey = self._getSubkey(receiveAddr, changeAddr, number)
        return subkey.address()

    def generateSegwitAddressFromXpub(self, receiveAddr:bool=True, changeAddr:bool=False, number=0):
        "For Ledger Xpubs that are actually used as ypubs. Alternative: Convert to ypub: https://jlopp.github.io/xpub-converter/"
        subkey = self._getSubkey(receiveAddr, changeAddr, number)
        #(#P2WPKH in P2SH / ypub) -->pycoin/key/BIP49Node.py
        script = self.Network.contract.for_p2pkh_wit(subkey.hash160(is_compressed=True))
        return self.Network.address.for_p2s(script)

    def generateNativeSegwitAddressFromXpub(self, receiveAddr:bool=True, changeAddr:bool=False, number=0):
        "For Ledger Xpubs that are actually used as ypubs"
        subkey = self._getSubkey(receiveAddr, changeAddr, number)
        #(#P2WPKH) -->pycoin/key/BIP84Node.py
        return self.Network.address.for_p2pkh_wit(subkey.hash160(is_compressed=True))
    
    def getNumTxForAddress(self, address, txData):
        pass

    def _getRawTxsForAddress(self, address, offset=None):
        print("Querying address {}".format(address))
        return blockcypher.get_address_details(address, coin_symbol=self.Network.symbol.lower())
    
    def _getRawTxDetails(self, txHash):
        #print("Querying Tx {}".format(txHash))
        return blockcypher.get_transaction_details(txHash, coin_symbol=self.Network.symbol.lower())

    def _query(self, url):
        '''
        Query the data. Takes a url as an input, calls the url, and returns the data
        
        :param url: the url to query
        :return: A dictionary of data.
        '''
        response = requests.get(url)
        print("Calling {}".format(url))
        if response.status_code != 200:
            raise BaseException("Wrong output")
        print(response.ok)
        print(response.status_code)
        data = response.json()
        time.sleep(10.5) #rate limit blockchain.info
        return data
    
    def _getTransactions(self, addressGenerator, offset=None, startTime=None, receiving=True):
        while(True):
            address = next(addressGenerator)
            data = self._getRawTxsForAddress(address, offset)
            num = data["n_tx"]
            if num == 0:
                break
            for elem in data["txrefs"]: #txs blockchain info
                txHash = elem["tx_hash"]
                change = satoshi_to_btc(elem["value"])
                # Check for out or inbound transaction
                isOut = elem["tx_output_n"] < 0
                isIn = elem["tx_output_n"] >= 0
                txFound = False

                if self.txList.hasTx(txHash):
                    t, idx = self.txList.fromTxHash(txHash)
                    txFound = True
                    #self.txList[idx] = t
                else:
                    # Fill new transaction object
                    t = Transaction()
                    t.datetime = elem["confirmed"]
                    t.txHash = txHash
                    
                if isIn and receiving:
                        t.posIn += Position(change, self.Network.symbol)
                if isOut:
                        t.posOut += Position(change, self.Network.symbol)
                if isIn and not(receiving):
                        t.posOut += Position(-change, self.Network.symbol)
                
                if not(txFound):
                    self.txList.append(t)


        return self.txList

    def _getAllTransactions(self, addressType:AddresType = AddresType.LEGACY):

        if addressType == AddresType.LEGACY:
            func = self.generateAddress

        elif addressType == AddresType.SEGWIT:
            func = self.generateSegwitAddressFromXpub
        
        elif addressType == AddresType.NATIVE_SEGWIT:
            func = self.generateNativeSegwitAddressFromXpub

        addressGenerator = self.generateAddresses(func, receiveAddr=True, changeAddr=False) 
        self._getTransactions(addressGenerator)
        addressGenerator = self.generateAddresses(func, receiveAddr=False, changeAddr=True) 
        self._getTransactions(addressGenerator, receiving=False)
        
        self.txList.sort(key=lambda r: r.datetime)


    def getTransactions(self, startTime=None, offset=None) -> 'TransactionList':
        #We test all 3 address types in case we have a ypub/zpub key as xpub key (as the Ledger software is doing)
        self._getAllTransactions(addressType=AddresType.LEGACY)
        time.sleep(5)
        self._getAllTransactions(addressType=AddresType.SEGWIT)
        time.sleep(5)
        self._getAllTransactions(addressType=AddresType.NATIVE_SEGWIT)

        #Sort
        self.txList.sort(key=lambda r: r.datetime)

        # Add fees only to outgoing transactions
        for tx in self.txList:
            if tx.isOutBound() and tx.fee.amount == 0:
                details = self._getRawTxDetails(tx.txHash)
                tx.fee = Fee( satoshi_to_btc(details["fees"]), self.Network.symbol)
            print(tx)

        return self.txList
        



#from btctools import Xpub
from cryptotools.BTC import Xpub, Address
import cryptotools
import requests
import datetime

#https://github.com/mcdallas/cryptotools
#https://bitcoin.stackexchange.com/questions/109981/derive-segwit-addresses-from-xpub-or-zpub-using-python
#https://bitcoin.stackexchange.com/questions/60565/is-it-possible-to-use-xpub-address-to-monitor-balance-from-previously-generated

#Learn: https://learnmeabitcoin.com/technical/derivation-paths


########################################

################ BITCOIN ################

#extended = Xpub.decode('xpub6C2zcaw8NCGbNwEo5KGPBYzNj6UpfspemmAfgsgxWTRvNPrXR6Sh9pvZvfNTjHaZ8mVPZjxMgi8RbCgecGSecU64agjh9uxQMXwPkxPf9bP')
extended = Xpub.decode('xpub6DBEa6eSZXgZ6GAkiyoWW5QEVKoU5oQ3geRyXfwmYTx75u5sMcB3LMEs5wSjKhPazrzFG1as4cu9Lgb53j1NRujueWqpn7GgmjkYJm9hs63')
print(extended.type)
#xpub6CEJ63Dk5YFhWhFraNVYz36VUhgQy1ed6j5gRDyJyJuWjM33KyDfS4hSWAa6CVUF7g3VtCqeEHs1TV1tgyYyWLvT4566DJYDbfG59DM3SGm #segwit
#xpub6DBEa6eSZXgZ6GAkiyoWW5QEVKoU5oQ3geRyXfwmYTx75u5sMcB3LMEs5wSjKhPazrzFG1as4cu9Lgb53j1NRujueWqpn7GgmjkYJm9hs63 #legacy
child = extended/0/0
#0/x = Receiving = 0 - Addresses that we will give out to people for receiving payments.
#1/x = Change = 1 - Addresses we use for sending change back to ourselves when we make transactions.

print(child)
print(child.key.to_address('P2WPKH')) #native segwite # bip84
print(child.key.to_address('P2PKH')) #Pay-to-Pubkey Hash - legacy  # bip44
#print(child.key.to_address('P2SH')) #segwit
print(child.key.to_address('P2WPKH-P2SH')) #segwit

curAddress = child.key.to_address('P2WPKH-P2SH')
address = Address(curAddress)
print(address.balance())
print(address.utxos)

##########################################

#1KCECbqEiHmdeLnPaBDc3J7h4upguV7M9S 



#Test:MULiubAyQSZNDgvvtj7q5D64vMbCCJfafP
print("###")

## doesnt work
def getbalance(address):
  response = requests.get("https://blockchain.info/rawaddr/%s" % address)
  if response.status_code != 200:
    return None
  return response.json()["final_balance"]
  
def listtransactions(address):
  response = requests.get("https://blockchain.info/rawaddr/%s" % address)
  if response.status_code != 200:
    return None
  for tx in response.json()["txs"]:
      #print(tx["time"])
      tx_timestr = datetime.datetime.fromtimestamp(tx["time"])
      tx_amount = satoshis_to_btc(tx["result"])
      print("Transaction (%s): %s %s" % (tx_timestr, "unknown type", tx_amount))
    
  return response.json()

def satoshis_to_btc(value):
  return float(float(value) / 10**8)


print(getbalance(curAddress))
listtransactions(curAddress)




####


import requests
import json

command = {
  "command": "findTransactions",
  "addresses": [
    "RVORZ9SIIP9RCYMREUIXXVPQIPHVCNPQ9HZWYKFWYWZRE9JQKG9REPKIASHUUECPSQO9JT9XNMVKWYGVA"
  ]
}

stringified = json.dumps(command)

headers = {
    'content-type': 'application/json',
    'X-IOTA-API-Version': '1'
}

request = requests.get(url="http://iota.org:14265", data=stringified, headers=headers)


jsonData = json.loads(request.json())

print(jsonData)


from pyCryptoTransactions.Transaction import Transaction, TransactionList
from .CSVImporter import CSVImporterDefault

from copy import deepcopy

class StakeTaxImporter(CSVImporterDefault):
    """Imports exported files from stake.tax (default format)"""
    def __init__(self, filename):
        super().__init__(filename)

        self.mapProperties = {"timestamp":"datetime", 
                              "tx_type":"category",
                              "taxable":None,
                              "received_amount": "posIn.amount",
                              "received_currency": "posIn.currency",
                              "sent_amount": "posOut.amount",
                              "sent_currency": "posOut.currency",
                              "fee": "fee.amount",
                              "fee_currency":"fee.currency",
                              "comment": "note",
                              "txid":"txHash",
                              "url": None,
                              "exchange":"network",
                              "wallet_address":"wallet" }

    def parseFile(self):
        self.checkFields()
        assert len(self.mapProperties) == len(self.df.columns), "Wrong number of columns in the dataframe"

        #Remove _blockchain from network name:
        self.df["exchange"] = self.df["exchange"].str.replace("_blockchain","")

        txList = TransactionList()
        
        t = Transaction()
        for index, row in self.df.iterrows():
            tmp = deepcopy(t)
            for key, val in self.mapProperties.items():
                if val == None:
                    continue
                if val.find('.')!=-1:
                    parent, child = val.split('.')
                    setattr(getattr(tmp, parent), child, row[key])
                else:
                    setattr(tmp,val,row[key])
            txList.append(tmp)
            del tmp

        return txList

    def checkFields(self):
        for key in self.mapProperties:
            if key not in self.df:
                print("Warning: {} is not included in file {}".format(key, self.filename))
    
    def getTransactions(self, time=None) -> TransactionList:
        #return super().getTransactions(address, time=time)
        return self.parseFile()




#https://midgard.thorchain.info/v2/actions?offset=0&limit=10&address=thor1khjg7et54q4zzef62ud0z8ps0ypnf69g5hfxjc

#https://thornode.thorchain.info/bank/balances/thor1khjg7et54q4zzef62ud0z8ps0ypnf69g5hfxjc

#https://midgard.thorchain.info/v2/member/thor1khjg7et54q4zzef62ud0z8ps0ypnf69g5hfxjc --> all pools

from datetime import datetime
import pytz as timezone
from decimal import Decimal
from copy import deepcopy

from pyCryptoTransactions.Transaction import Position, Transaction,TransactionList, Fee
from pyCryptoTransactions.Importer import Importer
import time
import datetime

class ThorchainImporter(Importer):
    def __init__(self, address):
        super().__init__()
        self._apiAdress = 'https://midgard.thorchain.info/v2/actions'
        self.balanceApiAdress = "https://thornode.thorchain.info/bank/balances/"
        self._initRequest()

        self.rawTxs = None
        self.address = address

        self.denominator = Decimal(int(1E8))
        self.chain = "thorchain"
        self.unit = "RUNE"
    
    def _getRawTransactions(self, offset=0, limit=10, **kwargs):
        values = {}
        values['limit'] = limit
        values['offset'] = offset
        values['address'] = self.address
        for k in kwargs:
            if kwargs[k] is not None:
                values[k] = kwargs[k]
        return self._request('/', **values)

    def _getAllRawTransactions(self):
        return NotImplemented

    def getTransactions(self, time=None) -> TransactionList:
        txsRaw = self._getRawTransactions() #todo:replace with getAllRawTransactions

        for tx in txsRaw["actions"]:
            time = datetime.datetime.fromtimestamp(int(tx["date"])/1E9,tz=timezone.utc)
            height = tx["height"]
            category = tx["type"]
            status = tx["status"] == "success"
            t = Transaction(dateTime=time, blockHeight=height, category=category)
            if "in" in tx:
                for txOut in tx["in"]: #thorchain api in means actually outgoing tx from us
                    if txOut["address"] == self.address:
                        t.fromAddress = self.address
                        for coin in txOut["coins"]:
                            tmp = deepcopy(t)
                            tmp.txHash = txOut["txID"]
                            tmp.posOut = Position(Decimal(coin["amount"])/self.denominator, coin["asset"].split(".")[1].split('-')[0])
                            if tmp.category == "addLiquidity":
                                tmp.fee = Fee(Decimal("0.02"),"RUNE")
                            self.txList.append(tmp)
            if "out" in tx:
                for txIn in tx["out"]: #thorchain api out means actually ingoging tx to us
                    if txIn["address"] == self.address:
                        t.toAddress = self.address
                        for coin in txIn["coins"]:
                            tmp = deepcopy(t)
                            tmp.txHash = txIn["txID"]
                            tmp.posIn = Position(Decimal(coin["amount"])/self.denominator, coin["asset"].split(".")[1].split('-')[0])
                            self.txList.append(tmp)
        return self.txList


from pyCryptoTransactions.Transaction import TransactionList
import pandas as pd

class ExporterDefault(object):

    def __init__(self, txData:TransactionList) -> None:
        super().__init__()
        self.txData = txData
        self.df = pd.DataFrame()

    def exportToCsv(self, outCsvFilename) -> None:
        self.df.to_csv(outCsvFilename)

    def exportToExcel(self, outExcelFilename) -> None:
        self.df.to_excel(outExcelFilename, index=False)



from requests import Request, Session
from requests.exceptions import ConnectionError, Timeout, TooManyRedirects
import json
from datetime import datetime
import time
#from django.utils import timezone
import pytz as timezone
from decimal import Decimal
from copy import deepcopy

from pyCryptoTransactions.Transaction import Position, Transaction,TransactionList, Fee
from pyCryptoTransactions.Importer import Importer

class BinanceChain(Importer):
    def __init__(self, address):
        super().__init__()
        self._apiAdress = 'https://dex.binance.org/api/v1'
        self._initRequest()
        self.rawTxs = None
        self.address = address
        self.denominator = Decimal(int(1E8))
        self.cancelledOrders = list()
    
    def _initRequest(self):
        headers = {'Accepts': 'application/json'}
        self.session = Session()
        self.session.headers.update(headers)
    
    def _request(self, path, **params):
        try:
            #print(params)
            response = self.session.get(self._apiAdress + path, params=params)
            #print(response.url)
            data = json.loads(response.text)
            if "ok" in data:
                if not(data["ok"]):
                    print(data)
                    raise("Error in the Request")
            if "code" in data:
                if data["code"] != 0:
                    print(data)
                    raise("Error in the request")
            if data: #data['ok'] and data["code"]==0:
                return data
            else:
                raise("Data failure:" + data['log'])
        except (ConnectionError, Timeout, TooManyRedirects) as e:
            print(e)
            raise("Connection Error")
    
    # Endpoint doesn't give multisend transactions!
    def _getRawTransactions(self, offset=0, limit=500, **kwargs):#offset=0, limit=500, startTime=None, endTime = int(round(time.time() * 1000))):
        values = {}
        values['address'] = self.address
        values['offset'] = offset
        values['limit'] = limit
        for k in kwargs:
            if kwargs[k] is not None:
                values[k] = kwargs[k]
        self._request('/transactions', **values)

    def getTransaction(self, txId):
        """ Get a transaction for a given tx ID """
        #https://docs.binance.org/api-reference/dex-api/paths.html#apiv1transactions
        args = {}
        args['format'] = 'json'
        return self._request('/tx/' + txId, **args)
    
    def getOrder(self, orderId):
        args = {}
        args["format"] = 'json'
        return self._request('/orders/' + orderId, **args)
    
    def getTrades(self, orderId, timestamp:int, isBuy:bool):
        args = {}
        args["format"] = 'json'
        args["start"] = timestamp -10
        args["end"] = timestamp + 10
        if isBuy:
            args["buyerOrderId"] = orderId
        else:
            args["sellerOrderId"] = orderId
        print(args)
        return self._request('/trades', **args)

    def getFees(self):
        return self._request('/fees')

    #def getTrades(self,offset=0, limit=500, **kwargs):
    #    values = {}
    #    values['offset'] = offset
    #    values['limit'] = limit
    #    for k in kwargs:
    #        if kwargs[k] is not None:
    #            values[k] = kwargs[k]
    #    self._request('/trades', **values)


    def getTransactionsPage(self,page=-1):
        """get Raw Tx Data for the given address from the binance explorer. 
           Note: page must be negative!
        """
        # Get a page also inlcuding multisend transactions
        #https: // github.com / binance - chain / java - sdk / issues / 30
        #https://explorer.binance.org/api/v1/txs?page=-1&rows=100&address=ADDRESS
        params = {'address':self.address, 'page':page, 'rows':100}
        tmpApiAddress = self._apiAdress
        self._apiAdress = "https://explorer.binance.org/api/v1"
        ret = self._request('/txs', **params)
        self._apiAdress = tmpApiAddress
        return ret

    def _getAllRawTxsFromTxPage(self):
        # Query first page
        page = -1
        dataTransactionPage = self.getTransactionsPage(page=page)
        numTxs = dataTransactionPage["txNums"]
        numTxsQueried = 100
        while numTxsQueried < numTxs:
            page -= 1
            tmp = self.getTransactionsPage(page=page)
            numTxsQueried += 100
            print(type(tmp))
            dataTransactionPage.update(tmp)
            #d.update(tmp)
            #dataTransactionPage += dict(dataTransactionPage.items() | tmp.items())
        self.rawTxs = dataTransactionPage
        #todo: remove all txs over a given minimum time

    # currently unused
    def _getAllMultiSendTransactionIds(self):
        multiSendTxIds = []
        for entry in self.rawTxs:
            if entry["hasChildren"] > 0:
                multiSendTxIds.append(entry["txHash"])

        return multiSendTxIds

    def _getAllTransactionIds(self):
        ids = []
        for entry in self.rawTxs["txArray"]:
            ids.append((entry["timeStamp"],entry["txHash"],entry["txFee"]))
        return ids
    
    #def manualImportTransactionFromID(self, txID):
    #    tx = self.getTransaction(txID)
    #    t = Transaction()
    #    t.txHash = result["hash"]
    #    t.datetime = datetime.fromtimestamp(timestamp / 1000, tz=timezone.utc)

    def getTransactions(self, startTime=None, offset=None) -> 'TransactionList':
        self._getAllRawTxsFromTxPage()
        for timestamp, id, feeValue in self._getAllTransactionIds():
            feeValue = Decimal(str(feeValue))
            result = self.getTransaction(id)

            t = Transaction(network="binanceChain")

            #only transactions of type "auth/StdTx" are supported
            if result["tx"]["type"] != "auth/StdTx":
                print("Unsupported transaction with id {}".format(id))
                continue

            t.txHash = result["hash"]
            t.datetime = datetime.fromtimestamp(timestamp / 1000, tz=timezone.utc)

            #print(result["tx"]["value"]["msg"])
            # Only take first one, typically there are no more , TODO: Check or do better :)
            if result["tx"]["value"]["msg"][0]["type"] == "cosmos-sdk/Send":
                transaction = result["tx"]["value"]["msg"][0]["value"]

                isMulitSend = self.isMultiSendTransaction(result)
                isOutTransaction = self.isOutTransaction(result)

                if isOutTransaction:
                    t.category = "Transfer Out" if not(isMulitSend) else "Stake"
                    if feeValue>0:
                        t.fee = Fee(feeValue,"BNB")
                    else:
                        t.fee = Fee(Decimal(37500)/self.denominator if not(isMulitSend) else Decimal(30000)/self.denominator, "BNB")
                    coins = transaction["outputs"][0]["coins"]
                    #fixedFees = self.getFees()["fixed_fee_params"]  # Todo query fixed fees
                else:
                    t.category = "Transfer In" if not (isMulitSend) else "Harvest"
                    coins = transaction["inputs"][0]["coins"]

                if not(isMulitSend) and result["tx"]["value"]["memo"].find("SWAP") != -1 or result["tx"]["value"]["memo"].find("OUTBOUND") != -1:
                    t.category = "Swap"

                if result["tx"]["value"]["memo"]:
                    t.note = "memo:" + result["tx"]["value"]["memo"]

                i = 0
                for coin in coins:
                    tmp = deepcopy(t)
                    if isOutTransaction:
                        tmp.posOut = Position(Decimal(coin["amount"])/self.denominator, coin["denom"].split("-")[0])
                    else:
                        tmp.posIn = Position(Decimal(coin["amount"])/self.denominator, coin["denom"].split("-")[0]) 
                    if i>0:
                        tmp.fee.amount = Decimal(0)
                    self.txList.append(tmp)
                    i+=1
                    # Hack: Put equivalent transaction for BEPSWAP
                    #if tmp["category"] == "Stake":
                    #    tmp2 = tmp.copy()
                    #    tmp2["amountIn"] = tmp2["amountOut"]
                    #    tmp2["currencyIn"] = tmp2["currencyOut"]
                    #    tmp2["amountOut"] = 0
                    #    tmp2["category"] = "LP"
                    #    tmp2["wallet"] = "Bepswap"
                    #    transactios.append(tmp2)

            elif result["tx"]["value"]["msg"][0]["type"] == "dex/NewOrder": #BUY
                t.note = "dex/New Order"
                t.category = "Trade"
                order = result["tx"]["value"]["msg"][0]["value"]
                t.orderId = order["id"]
                orderInfo = self.getOrder(t.orderId)
                pair = order["symbol"].split('_')

                trades = self.getTrades(timestamp=int(t.datetime.timestamp()*1E3), orderId=t.orderId, isBuy=orderInfo["side"]==1)
                if order["ordertype"] == 2 and order["side"]==2: #sell
                    posIn = Position(Decimal(0),pair[1].split('-')[0])
                    posOut = Position(Decimal(0),pair[0].split('-')[0])
                    fee = Fee(0, "BNB")
                    for trade in trades["trade"]:
                        posIn += Position( Decimal(trade["price"].rstrip("0")) * Decimal(trade["quantity"].rstrip("0")), pair[1].split('-')[0])
                        posOut += Position( Decimal(trade["quantity"].rstrip("0")), pair[0].split('-')[0])
                    fee = Fee( Decimal(trade["sellFee"].split(";")[0].split("BNB:")[1]), "BNB")
                elif order["ordertype"] == 2 and order["side"]==1: #buy
                    posIn = Position(Decimal(0),pair[0].split('-')[0])
                    posOut = Position(Decimal(0),pair[1].split('-')[0])
                    for trade in trades["trade"]:
                        posIn += Position(  Decimal(trade["quantity"].rstrip("0")), pair[0].split('-')[0])
                        posOut += Position(Decimal(trade["price"].rstrip("0")) * Decimal(trade["quantity"].rstrip("0")), pair[1].split('-')[0])
                    fee = Fee( Decimal(trade["buyFee"].split(";")[0].split("BNB:")[1]), "BNB")

                t.posIn = posIn
                t.posOut = posOut
                t.fee = fee
                    
                #orderPrice1 = Decimal(order["price"])/self.denominator
                #orderQuantity1 = Decimal(order["quantity"])/self.denominator

               #print("{} == P1:{} Q1:{}, P2:{}, Q2:{} ({})".format(t.orderId, orderPrice1,orderQuantity1,orderPrice,orderQuantity,pair))

                if t.orderId in self.cancelledOrders or orderInfo["status"]=="Canceled":
                    continue
                #if order["ordertype"] == 2 and order["side"]==2: #sell
                #    t.posIn = Position(orderPrice * orderQuantity, pair[1].split('-')[0])
                #    t.posOut = Position(orderQuantity, pair[0].split('-')[0])
                #elif order["ordertype"] == 2 and order["side"]==1: #buy
                #    t.posIn = Position(orderQuantity, pair[0].split('-')[0])
                #    t.posOut = Position(orderPrice * orderQuantity, pair[1].split('-')[0])
                
                #t.fee = Fee(0, "BNB")
                t.price = Decimal(order["price"])/self.denominator
                t.tradingPair = (pair[0].split('-')[0], pair[1].split('-')[0])
                
                self.txList.append(t)

            elif result["tx"]["value"]["msg"][0]["type"] == "bridge/TransferOutMsg":
                # Swap to bsc network chain
                t.note = "Atomic Swap to BSC Chain"
                t.category = "Swap"
                t.posOut = Position(Decimal(result["tx"]["value"]["msg"][0]["value"]["amount"]["amount"]) / self.denominator, result["tx"]["value"]["msg"][0]["value"]["amount"]["denom"].split('-')[0])
                t.fee = Fee(Decimal("0.0002"), "BNB")
                self.txList.append(t)
            elif result["tx"]["value"]["msg"][0]["type"] == "dex/CancelOrder":
                # we must find orders that match this order id or prevent new ones getting added
                order = result["tx"]["value"]["msg"][0]["value"]
                orderId = order["refid"]
                if self.txList.fromOrderId(orderId) is not None:
                    self.txList.remove(self.txList.fromOrderId(orderId))
                else:
                    self.cancelledOrders.append(orderId)
                t.category = "Loss"
                t.fee = Fee(Decimal("0.00001"),"BNB")
                self.txList.append(t)
            else:
                raise("Unknown type: %s" % result["tx"]["value"]["msg"][0]["type"])

        return self.txList

    def isMultiSendTransaction(self, txResult):
        value = txResult["tx"]["value"]["msg"][0]["value"]
        if "inputs" in value and len(value["inputs"][0]["coins"]) > 1:
            return True
        if "outputs" in "value" and len(value["outputs"][0]["coins"]) > 1:
            return True
        return False

    def isOutTransaction(self, txResult):
        value = txResult["tx"]["value"]["msg"][0]["value"]
        if "inputs" in value and value["inputs"][0]["address"] == self.address:
            return True
        elif "outputs" in value and value["outputs"][0]["address"] == self.address:
            return False
        else:
            raise("Error: Input address not found in transacion entry!")

    #def get(self, address):
    #    # try to grab it all
    #    trades = self.getTrades()
    #    transactions = self._getRawTransactions(address=address, startTime=1609894800000, endTime=int(round(time.time() * 1000)))
    #    multiSendTransactions = []
    #    multiSendTransactionIds = self._getAllMultiSendTransactionIds(address=address)
    #    for id in multiSendTransactions:
    #        multiSendTransactions.append(self.getTransaction(id))

        # Now put it all together:
        # tbd








from .CoinbaseAPIKeyAuthentication import CoinbaseAPIKeyAuthentication
from .CoinbaseOAuthAuthentication import CoinbaseOAuthAuthentication
from .error import CoinbaseAPIException

import hashlib
import hmac
import json
import requests
import time
import urllib.parse
class CoinbaseRPC(object):
    '''
    Abstracts functionality for executing remote procedure calls.
    '''
    
    COINBASE_API_BASE = 'https://api.coinbase.com/'
    API_VERSION = 2
    COINBASE_API = COINBASE_API_BASE + 'v' + str(API_VERSION)
    def __init__(self, authentication, nonce=None):
        self.__authentication = authentication
        self.__nonce = None


    def request(self, method, url, params=None):
        url = self.COINBASE_API + url

        method = method.lower()
        if method == 'get' or method == 'delete':
            if params is not None:
                url += '?' + urllib.parse.urlencode(params)
        else:
            params = json.dumps(params)

        headers = {
            'Content-Type': 'application/json',
            'User-Agent': 'CoinbasePython3/v2'
        }

        auth = self.__authentication.get_data()

        if isinstance(self.__authentication, CoinbaseOAuthAuthentication):
            headers['Authorization'] = 'Bearer ' + auth['access_token']
        elif isinstance(self.__authentication, CoinbaseAPIKeyAuthentication):
            if self.__nonce is None:
                self.__nonce = int(time.time() * 1e6)
            message = str(self.__nonce) + url

            if method == 'post' or method == 'put':
                if params is not None:
                    message += params

            signature = \
                hmac.new(
                    auth['api_secret'].encode(),
                    message.encode(),
                    hashlib.sha256) \
                .hexdigest()

            headers['CB-ACCESS-KEY'] = auth['api_key']
            headers['CB-ACCESS-SIGN'] = signature
            headers['CB-ACCESS-TIMESTAMP'] = str(self.__nonce)
            headers['CB-VERSION'] = '2021-12-25'
            headers['Accept'] = 'application/json'
        else:
            raise CoinbaseAPIException('Invalid authentication mechanism')

        #debug: print(f"{method} {url} {headers}")
        if method == 'get':
            request = requests.get(url, headers=headers)
        elif method == 'delete':
            request = requests.delete(url, headers=headers)
        elif method == 'post':
            request = requests.post(url, data=params, headers=headers)
        elif method == 'put':
            request = requests.put(url, data=params, headers=headers)

        if request.status_code != 200:
            raise CoinbaseAPIException(
                    'Status Code ' + str(request.status_code),
                    request.status_code,
                    request.content.decode())

        content = json.loads(request.content.decode())
        if content is None or len(content) == 0:
            raise CoinbaseAPIException(
                    'Invalid response body',
                    request.status_code,
                    request.content.decode())

        if 'error' in content:
            raise CoinbaseAPIException(
                    content['error'],
                    request.status_code,
                    request.content.decode())

        if 'errors' in content:
            raise CoinbaseAPIException(
                    ', '.join(content['errors']),
                    request.status_code,
                    request.content.decode())

        return content



from pyCryptoTransactions.networks.cosmos.cosmos import CosmosChain

class PersistenceCHain(CosmosChain):

    def __init__(self, address):
        super().__init__(address)

        self._apiAdress = "https://api-persistence.cosmostation.io/v1/"
        self.chain = "osmosis-1"
        self.ibcTokenApiAddress = "https://api-utility.cosmostation.io/v1//ibc/tokens/osmosis-1"
        self.unit = "XPRT"
        self._getAllIBCTokens()
    
    @property
    def pathToMemo(self):
        return ["data","tx","body","memo"]
    
    @property
    def pathToFee(self):
        return ["data","tx","auth_info","fee","amount",0,"amount"]
    
    @property
    def pathToFeeCurrency(self):
        return ["data","tx","auth_info","fee","amount",0,"denom"]


from .CoinbaseAuthentication import CoinbaseAuthentication
class CoinbaseOAuthAuthentication(CoinbaseAuthentication):
    '''
    Implements the authentication mechanism that uses an access token and a
    refresh token.
    '''

    def __init__(self, access_token, refresh_token):
        self.access_token = access_token
        self.refresh_token = refresh_token


    def get_data(self):
        return {
            'access_token': self.access_token,
            'refresh_token': self.refresh_token
        }



from cryptotools.BTC import Xpub
import cryptotools
import requests
import datetime

#Change bitcoin network settings to litecoin
#https://bitcoin.stackexchange.com/questions/62781/litecoin-constants-and-prefixes/80355#80355
cryptotools.BTC.network.main["extended_pub"]["P2PKH"] = b'\x01\x9d\xa4\x62' #0x019DA462 #Ltub
cryptotools.BTC.network.main["hrp"] = 'ltc' #0x019DA462 #Ltub
cryptotools.BTC.network.main["keyhash"] = b'\x30'
cryptotools.BTC.network.main["scripthash"] = b'\x32' #new M-address
cryptotools.BTC.network.main["wif"] = b'\x36' #not important?

segwitXpub = 'Ltub2ZYYmNuYZRAyCFNp8LAPEwxnRG9jkx4qmmPDwTfQNe19JgURHxAL1WjoMP24ZP8fxZafCTd8SBR9acr9KWjWmM3rtyFsRfQf3dk3cK2hgV5'
legacyXpub = 'Ltub2Z1YoRfU7gavCJ5ChwGNAnmcpDK35wzCRAB2mstyVjV7ShQGnizKTyL6kFNhK4KNuWypAYvAS534LFowXrFRPLQRacRBemKrrAWCzuPEe4o'
extended = Xpub.decode(legacyXpub)
child = extended/0/0
print(child.key.to_address('P2WPKH')) #native segwite # bip84
print(child.key.to_address('P2PKH')) #Pay-to-Pubkey Hash - legacy  # bip44
print(child.key.to_address('P2WPKH-P2SH')) #segwit

#LbNrLRMKNG68zHeNnDdMotK8Vr2jLz3bej
#add from legacy xpub wrong (decode assumes segwit?)



from cryptotools.BTC import Xprv
m = Xprv.from_mnemonic('bench auction print immense castle slide tip pill evolve abandon canvas ugly')
M = m.to_xpub()
print(M)
print(M.encode())
print(M.address('P2PKH'))
print((m/0./0).to_xpub())
print((m/0./0).address('P2PKH'))  # bip44)
print((m/0./0).to_xpub().encode())
#(m/123/456).to_xpub()
#(m/44./0./0./0/0).address('P2PKH')


import hashlib
from pycoin import ecdsa, encoding
import os
import codecs
for i in range(10):
    rand = codecs.encode(os.urandom(32), 'hex').decode()
    secret_exponent= int('0x'+rand, 0)
    print ('WIF: ' + encoding.secret_exponent_to_wif(secret_exponent, compressed=False))
    public_pair = ecdsa.public_pair_for_secret_exponent(ecdsa.secp256k1.generator_secp256k1, secret_exponent)
    hash160 = encoding.public_pair_to_hash160_sec(public_pair, compressed=True)
    print('Bitcoin address: %s' % encoding.hash160_sec_to_bitcoin_address(hash160))


#!/usr/bin/python

import pycoin.key
import sys
import requests
import json
import datetime
from pycoin.key.Key import Key

def getbalance(address):
  response = requests.get("https://bitaps.com/api/address/%s" % address)
  if response.status_code != 200:
    return None
  return response.json()
  
def listtransactions(address):
  response = requests.get("https://bitaps.com/api/address/transactions/%s" % address)
  if response.status_code != 200:
    return None
  return response.json()

def satoshis_to_btc(value):
  return float(float(value) / 10**8)

def probe_used_addresses(xpub, account_type):
  xpub_subkey = xpub.subkey(account_type)
  index = 0
  while True:
    addr = xpub_subkey.subkey(index).bitcoin_address()
    print("Type %d (%d): %s" % (account_type, index, addr))
    
    # Report balance data
    balancedata = getbalance(addr)
    balance = balancedata['confirmed_balance']
    balance_btc = satoshis_to_btc(balance)
    print("  Balance: %s" % balance_btc)
    
    # List transactions, if any...
    transactions = listtransactions(addr)
    if transactions != None:
      for tx in transactions:
        tx_time = datetime.datetime.fromtimestamp(float(tx[0]))
        tx_hash = tx[1]
        tx_data = tx[2]
        tx_type = tx[3]
        tx_status = tx[4]
        tx_confirmations = tx[5]
        tx_block = tx[6]
        tx_amount = satoshis_to_btc(tx[7])
        tx_timestr = tx_time.isoformat()
        print("    Transaction (%s): %s %s" % (tx_timestr, tx_type, tx_amount))

    print("")
    
    if transactions == None:
        break
    index += 1

def main():
  from pycoin.symbols.btc import network
  key = network.parse.bip32("xprv9s21ZrQH143K31AgNK5pyVvW23gHnkBq2wh5aEk6g1s496M"
      "8ZMjxncCKZKgb5jZoY5eSJMJ2Vbyvi2hbmQnCuHBujZ2WXGTux1X2k9Krdtq")
  print(key.hwif(as_private=1))
  print(key.hwif())
  print(key.wif())
  print(key.sec())
  print(key.address())

  xpub = Key.from_text(sys.argv[1])
  probe_used_addresses(xpub, 0)
  probe_used_addresses(xpub, 1)

main()
